/*
 * Transdisk V4.2x
 * Copyright 1995-97 Bernd Schmidt, Marcus Sundberg, Stefan Ropke,
 *                   Rodney Hester, Joanne Dow
 *
 * 2003-05-23: XDF mode added by Karsten Sperling
 *
 * Use DICE and 2.0 includes or above to compile. SAS C should work too.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <strings.h>

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <devices/trackdisk.h>

#include <clib/alib_protos.h>
#include <clib/exec_protos.h>

static void usage(void);

static void usage(void)
{
    fprintf(stderr, "Usage: transdisk options\n");
    fprintf(stderr, "Recognized options:\n");
    fprintf(stderr, "-h: Assume device is high density\n");
    fprintf(stderr, "-x: Use XDF file format (xon/xoff-safe)\n");
    fprintf(stderr, "-d device unit: Use this device instead of DF0:\n");
    fprintf(stderr, "-s n: Begin transfer at track n\n");
    fprintf(stderr, "-e n: End transfer at track n\n");
    fprintf(stderr, "-w filename: writes the ADF file <filename> to disk\n\n");
    fprintf(stderr, "Example:\n");
    fprintf(stderr, "transdisk >RAM:df1.adf.1 -d trackdisk 1 -s 0 -e 39\n");
    fprintf(stderr, "transfers the first half of the floppy in DF1: into\n");
    fprintf(stderr, "a file in the RAM disk.\n");
    fprintf(stderr, "Or:\n");
    fprintf(stderr, "transdisk -w test.adf\n");
    fprintf(stderr, "writes the ADF-file test.adf to the disk in df0:\n");
}

#define XON  0x11
#define XOFF 0x13
#define ESC  0x1b

int main(int argc, char **argv)
{
    char *filename, *openMode="rb";
    FILE *ADFFile;
    int write=0;
    int xdf=0;
    struct IOStdReq *ioreq;
    struct MsgPort *port;

    UBYTE *buffer;
    char devicebuf[256];
    char *devicename = "trackdisk.device";
    char devicenum = 0;
    int i;
    int starttr = 0, endtr = 79;
    int sectors = 11;

    for (i = 1; i < argc;) {
	if (argv[i][0] != '-' || argv[i][2] != 0) {
	    usage();
	    exit(1);
	}
	switch (argv[i][1]) {
	 case 'h':
	    sectors = 22;
	    i++;
	    break;
        case 'x':
            xdf = 1;
            i++;
            break;
	case 'd':
	    if (i+2 >= argc) {
		usage();
		exit(1);
	    }
	    devicenum = atoi(argv[i+2]);
	    sprintf(devicebuf, "%s.device", argv[i+1]);
	    devicename = devicebuf;
	    i += 3;
	    break;
	case 's':
	    if (i+1 >= argc) {
		usage();
		exit(1);
	    }
	    starttr = atoi(argv[i+1]);
	    i += 2;
	    break;
	case 'e':
	    if (i+1 >= argc) {
		usage();
		exit(1);
	    }
	    endtr = atoi(argv[i+1]);
	    i += 2;
	    break;
	case 'w':
	    if (i+1 >= argc) {
		usage();
		exit(1);
	    }
	    filename=argv[i+1];
	    write=1;
	    i += 2;
	    break;
	default:
	    usage();
	    exit(1);
	}
    }
    fprintf(stderr,"Using %s unit %d\n", devicename, devicenum);
    fprintf(stderr,"Tracks are %d sectors\n", sectors);
    fprintf(stderr,"First track %d, last track %d\n", starttr, endtr);

    buffer = AllocMem(xdf ? 1024 : 512, MEMF_CHIP);
    if (write) {
	ADFFile = fopen(filename,openMode);

	if (!ADFFile) {
	    fprintf(stderr,"Error while opening input file\n");
	    exit (1);
	}
    }
    
    port = CreatePort(0, 0);
    if (port) {
	ioreq = CreateStdIO(port);
	if (ioreq) {
	    if (OpenDevice(devicename, devicenum, (struct IORequest *) ioreq, 0) == 0) {
		int tr, sec;

		ioreq->io_Command = write ? CMD_WRITE : CMD_READ;
		ioreq->io_Length = 512;
		ioreq->io_Data = buffer;
		for (tr = starttr*2; tr < (endtr+1)*2; tr++) {
		    fprintf(stderr,"Track: %d\r",tr/2);
		    for (sec = 0; sec < sectors; sec++) {
			fflush(stderr);
			if (write) {
                            if (xdf) {
                                UBYTE *out = buffer, *in = buffer+512;
                                int todo = 512; /* bytes needed in output buffer */
                                while (todo > 0) {
                                    if (fread(in, sizeof(UBYTE), todo, ADFFile) < todo) {
                                        fprintf(stderr, "Error: XDF file too short?\n");
                                        exit(1);
                                    }
                                    /* Include left-over byte into the buffer */
                                    todo += in - (buffer+512);
                                    in = buffer+512;
                                    /* Unescape. todo = bytes left in input buffer */
                                    while (todo-- > 0) {
                                        UBYTE b = *(in++);
                                        if (b != ESC) *(out++) = b;
                                        else {
                                            if (todo-- > 0) {
                                                switch (*(in++)) {
                                                case 0:
                                                    *(out++) = XON;
                                                    break;
                                                case 1:
                                                    *(out++) = XOFF;
                                                    break;
                                                case ESC:
                                                    *(out++) = ESC;
                                                    break;
                                                default:
                                                    fprintf(stderr, "Error: Illegal XDF escape\n");
                                                    exit(1);
                                                }
                                            }
                                            else { /* buffer ends in ESC, set flag */
                                                todo = 1;
                                                break;
                                            }
                                        }
                                    }
                                    in = buffer+512;
                                    if (todo > 0) *(in++) = ESC; /* keep the ESC for next loop */
                                    
                                    todo = buffer+512-out;
                                }
                            }
                            else if (fread(buffer, sizeof(UBYTE), 512, ADFFile) < 512) {
				fprintf(stderr, "Error: ADF file too short?\n");
				exit(1);
			    }
                        }
			ioreq->io_Offset = 512 * (tr * sectors + sec);
			DoIO( (struct IORequest *) ioreq);
			if (!write) {
                            if (xdf) {
                                /* Escape xon/xoff/esc in output data */
                                UBYTE *in = buffer+512, *out = buffer+1024;
                                while (--in >= buffer) {
                                    switch (*in) {
                                    case XON:
                                        *--out = 0;
                                        *--out = ESC;
                                        break;
                                    case XOFF:
                                        *--out = 1;
                                        *--out = ESC;
                                        break;
                                    case ESC:
                                        *--out = ESC;
                                        *--out = ESC;
                                        break;
                                    default:
                                        *--out = *in;
                                    }
                                }
                                fwrite(out, sizeof(UBYTE), buffer+1024-out, stdout);
                            }
                            else fwrite(buffer, sizeof(UBYTE), 512, stdout);
                        }
		    }
		}
		if (write) {		/* Make sure the last track is written to disk */
		    ioreq->io_Command = CMD_UPDATE;
		    DoIO( (struct IORequest *) ioreq);
		}
		ioreq->io_Command = TD_MOTOR;	/* Turn Disk-motor off */
		ioreq->io_Length = 0;
		DoIO( (struct IORequest *) ioreq);
		CloseDevice( (struct IORequest *) ioreq);
	    } else
		fprintf(stderr,"Unable to open %s unit %d\n", devicename, devicenum);
	    DeleteStdIO(ioreq);
	}
	DeletePort(port);
    }
    fprintf(stderr,"\n");
    FreeMem(buffer, xdf ? 1024 : 512);
    if (write)
	fclose (ADFFile);
    return 0;
}
